"""cmemc constants"""

NAMESPACES = {
    "void": "http://rdfs.org/ns/void#",
    "di": "https://vocab.eccenca.com/di/",
    "shui": "https://vocab.eccenca.com/shui/",
    "dsm": "https://vocab.eccenca.com/dsm/",
    "owl": "http://www.w3.org/2002/07/owl#",
    "sd": "http://www.w3.org/ns/sparql-service-description#",
}

NS_ACL = "http://eccenca.com/ac/"
NS_USER = "http://eccenca.com/"
NS_GROUP = "http://eccenca.com/"
